from .blending import *  # noqa: F401, F403
from .combing import *  # noqa: F401, F403
from .funcs import *  # noqa: F401, F403
from .ivtc import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403



from __future__ import annotations

from vstools import core, vs

__all__ = [
    'telecine_patterns',
]


def telecine_patterns(clipa: vs.VideoNode, clipb: vs.VideoNode, length: int = 5) -> list[vs.VideoNode]:
    a_select = [clipa.std.SelectEvery(length, i) for i in range(length)]
    b_select = [clipb.std.SelectEvery(length, i) for i in range(length)]

    return [
        core.std.Interleave([
            (b_select if i == j else a_select)[j] for j in range(length)
        ]) for i in range(length)
    ]



from __future__ import annotations

import warnings
from typing import Any

from vstools import DependencyNotFoundError, FieldBased, FieldBasedT, InvalidFramerateError, check_variable, core, vs

__all__ = [
    'pulldown_credits'
]


def pulldown_credits(
    clip: vs.VideoNode, frame_ref: int, tff: bool | FieldBasedT | None = None,
    interlaced: bool = True, dec: bool | None = None,
    bob_clip: vs.VideoNode | None = None, qtgmc_args: dict[str, Any] | None = None
) -> vs.VideoNode:
    """
    Deinterlacing function for interlaced credits (60i/30p) on top of telecined video (24p).

    This is a combination of havsfunc's dec_txt60mc, ivtc_txt30mc, and ivtc_txt60mc functions.
    The credits are interpolated and decimated to match the output clip.

    The function assumes you're passing a telecined clip (that's native 24p).
    If your clip is already fieldmatched, decimation will automatically be enabled unless set it to False.
    Likewise, if your credits are 30p (as opposed to 60i), you should set `interlaced` to False.

    The recommended way to use this filter is to trim out the area with interlaced credits,
    apply this function, and `vstools.insert_clip` the clip back into a properly IVTC'd clip.
    Alternatively, use `muvsfunc.VFRSplice` to splice the clip back in if you're dealing with a VFR clip.

    :param clip:                    Clip to process. Framerate must be 30000/1001.
    :param frame_ref:               First frame in the pattern. Expected pattern is ABBCD,
                                    except for when ``dec`` is enabled, in which case it's AABCD.
    :param tff:                     Top-field-first. `False` sets it to Bottom-Field-First.
    :param interlaced:              60i credits. Set to false for 30p credits.
    :param dec:                     Decimate input clip as opposed to IVTC.
                                    Automatically enabled if certain fieldmatching props are found.
                                    Can be forcibly disabled by setting it to `False`.
    :param bob_clip:                Custom bobbed clip. If `None`, uses a QTGMC clip.
                                    Framerate must be 60000/1001.
    :param qtgmc_args:              Arguments to pass on to QTGMC.
                                    Accepts any parameter except for FPSDivisor and TFF.

    :return:                        IVTC'd/decimated clip with credits pulled down to 24p.

    :raises ModuleNotFoundError:    Dependencies are missing.
    :raises ValueError:             Clip does not have a framerate of 30000/1001 (29.97).
    :raises TopFieldFirstError:     No automatic ``tff`` can be determined.
    :raises InvalidFramerateError:  Bobbed clip does not have a framerate of 60000/1001 (59.94)
    """

    try:
        from havsfunc import QTGMC  # type: ignore[import]
    except ModuleNotFoundError:
        raise DependencyNotFoundError(pulldown_credits, 'havsfunc')

    try:
        from vsdenoise import prefilter_to_full_range
    except ModuleNotFoundError:
        from havsfunc import DitherLumaRebuild as prefilter_to_full_range  # type: ignore
        warnings.warn("pulldown_credits: missing dependency `vsdenoise`!", ImportWarning)

    assert check_variable(clip, "pulldown_credits")

    InvalidFramerateError.check(pulldown_credits, clip, (30000, 1001))

    tff = FieldBased.from_param(tff, pulldown_credits) or FieldBased.from_video(clip, True)
    clip = FieldBased.ensure_presence(clip, tff)

    qtgmc_kwargs = dict[str, Any](
        SourceMatch=3, Lossless=2, TR0=2, TR1=2, TR2=3, Preset="Placebo"
    ) | (qtgmc_args or {}) | dict(FPSDivisor=1, TFF=tff.field)

    if dec is not False:  # Automatically enable dec unless set to False
        dec = any(x in clip.get_frame(0).props for x in {"VFMMatch", "TFMMatch"})

        if dec:
            warnings.warn("pulldown_credits: 'Fieldmatched clip passed to function! "
                          "dec is set to `True`. If you want to disable this, set `dec=False`!'")

    # motion vector and other values
    field_ref = frame_ref * 2
    frame_ref %= 5
    invpos = (5 - field_ref) % 5

    offset = [0, 0, -1, 1, 1][frame_ref]
    pattern = [0, 1, 0, 0, 1][frame_ref]
    direction = [-1, -1, 1, 1, 1][frame_ref]

    blksize = 16 if clip.width > 1024 or clip.height > 576 else 8
    overlap = blksize // 2

    ivtc_fps = dict(fpsnum=24000, fpsden=1001)
    ivtc_fps_div = dict(fpsnum=12000, fpsden=1001)

    # Bobbed clip
    bobbed = bob_clip or QTGMC(clip, **qtgmc_kwargs)

    InvalidFramerateError.check(pulldown_credits, bobbed, (60000, 1001))

    if interlaced:  # 60i credits. Start of ABBCD
        if dec:  # Decimate the clip instead of properly IVTC
            clean = bobbed.std.SelectEvery(5, [4 - invpos])

            if invpos > 2:
                jitter = core.std.AssumeFPS(
                    bobbed[0] * 2 + bobbed.std.SelectEvery(5, [6 - invpos, 7 - invpos]),
                    **ivtc_fps)  # type:ignore[arg-type]
            elif invpos > 1:
                jitter = core.std.AssumeFPS(
                    bobbed[0] + bobbed.std.SelectEvery(5, [2 - invpos, 6 - invpos]),
                    **ivtc_fps)  # type:ignore[arg-type]
            else:
                jitter = bobbed.std.SelectEvery(5, [1 - invpos, 2 - invpos])
        else:  # Properly IVTC
            if invpos > 1:
                clean = core.std.AssumeFPS(bobbed[0] + bobbed.std.SelectEvery(5, [6 - invpos]),
                                           **ivtc_fps_div)  # type:ignore[arg-type]
            else:
                clean = bobbed.std.SelectEvery(5, [1 - invpos])

            if invpos > 3:
                jitter = core.std.AssumeFPS(bobbed[0] + bobbed.std.SelectEvery(5, [4 - invpos, 8 - invpos]),
                                            **ivtc_fps)  # type:ignore[arg-type]
            else:
                jitter = bobbed.std.SelectEvery(5, [3 - invpos, 4 - invpos])

        jsup_pre = prefilter_to_full_range(jitter, 1.0).mv.Super(pel=2)
        jsup = jitter.mv.Super(pel=2, levels=1)
        vect_f = jsup_pre.mv.Analyse(blksize=blksize, isb=False, delta=1, overlap=overlap)
        vect_b = jsup_pre.mv.Analyse(blksize=blksize, isb=True, delta=1, overlap=overlap)
        comp = core.mv.FlowInter(jitter, jsup, vect_b, vect_f)
        out = core.std.Interleave([comp[::2], clean] if dec else [clean, comp[::2]])
        offs = 3 if dec else 2
        return out[invpos // offs:]
    else:  # 30i credits
        if pattern == 0:
            if offset == -1:
                c1 = core.std.AssumeFPS(bobbed[0] + bobbed.std.SelectEvery(
                    10, [2 + offset, 7 + offset, 5 + offset, 10 + offset]), **ivtc_fps)  # type:ignore[arg-type]
            else:
                c1 = bobbed.std.SelectEvery(10, [offset, 2 + offset, 7 + offset, 5 + offset])

            if offset == 1:
                c2 = core.std.Interleave([
                    bobbed.std.SelectEvery(10, [4]),
                    bobbed.std.SelectEvery(10, [5]),
                    bobbed[10:].std.SelectEvery(10, [0]),
                    bobbed.std.SelectEvery(10, [9])
                ])
            else:
                c2 = bobbed.std.SelectEvery(10, [3 + offset, 4 + offset, 9 + offset, 8 + offset])
        else:
            if offset == 1:
                c1 = core.std.Interleave([
                    bobbed.std.SelectEvery(10, [3]),
                    bobbed.std.SelectEvery(10, [5]),
                    bobbed[10:].std.SelectEvery(10, [0]),
                    bobbed.std.SelectEvery(10, [8])
                ])
            else:
                c1 = bobbed.std.SelectEvery(10, [2 + offset, 4 + offset, 9 + offset, 7 + offset])

            if offset == -1:
                c2 = core.std.AssumeFPS(bobbed[0] + bobbed.std.SelectEvery(
                    10, [1 + offset, 6 + offset, 5 + offset, 10 + offset]), **ivtc_fps)  # type:ignore[arg-type]
            else:
                c2 = bobbed.std.SelectEvery(10, [offset, 1 + offset, 6 + offset, 5 + offset])

        super1_pre = prefilter_to_full_range(c1, 1.0).mv.Super(pel=2)
        super1 = c1.mv.Super(pel=2, levels=1)
        vect_f1 = super1_pre.mv.Analyse(blksize=blksize, isb=False, delta=1, overlap=overlap)
        vect_b1 = super1_pre.mv.Analyse(blksize=blksize, isb=True, delta=1, overlap=overlap)
        fix1 = c1.mv.FlowInter(super1, vect_b1, vect_f1, time=50 + direction * 25).std.SelectEvery(4, [0, 2])

        super2_pre = prefilter_to_full_range(c2, 1.0).mv.Super(pel=2)
        super2 = c2.mv.Super(pel=2, levels=1)
        vect_f2 = super2_pre.mv.Analyse(blksize=blksize, isb=False, delta=1, overlap=overlap)
        vect_b2 = super2_pre.mv.Analyse(blksize=blksize, isb=True, delta=1, overlap=overlap)
        fix2 = c2.mv.FlowInter(super2, vect_b2, vect_f2).std.SelectEvery(4, [0, 2])

        return core.std.Interleave([fix1, fix2] if pattern == 0 else [fix2, fix1])



from __future__ import annotations

import gc
import os
import sys
from pathlib import Path
from typing import Any

from vstools import (
    CustomEnum, CustomTypeError, FieldBased, FieldBasedT, InvalidFramerateError, VSFunctionKwArgs, clip_async_render,
    core, join, vs
)

from .blending import deblend

__all__ = [
    'IVTCycles',
    'sivtc', 'jivtc',
    'tivtc_2pass', 'tivtc_vfr'
]


class IVTCycles(list[int], CustomEnum):
    cycle_10 = [[0, 3, 6, 8], [0, 2, 5, 8], [0, 2, 4, 7], [2, 4, 6, 9], [1, 4, 6, 8]]
    cycle_08 = [[0, 3, 4, 6], [0, 2, 5, 6], [0, 2, 4, 7], [0, 2, 4, 7], [1, 2, 4, 6]]
    cycle_05 = [[0, 1, 3, 4], [0, 1, 2, 4], [0, 1, 2, 3], [1, 2, 3, 4], [0, 2, 3, 4]]

    @property
    def pattern_length(self) -> int:
        return int(self._name_[6:])

    @property
    def length(self) -> int:
        return len(self.value)

    def decimate(self, clip: vs.VideoNode, pattern: int = 0) -> vs.VideoNode:
        assert 0 <= pattern < self.length
        return clip.std.SelectEvery(self.pattern_length, self.value[pattern])


def sivtc(clip: vs.VideoNode, pattern: int = 0, tff: bool | FieldBasedT = True) -> vs.VideoNode:
    """
    Simplest form of a fieldmatching function.

    This is essentially a stripped-down JIVTC offering JUST the basic fieldmatching and decimation part.
    As such, you may need to combine multiple instances if patterns change throughout the clip.

    :param clip:        Clip to process.
    :param pattern:     First frame of any clean-combed-combed-clean-clean sequence.
    :param tff:         Top-Field-First.

    :return:            IVTC'd clip.
    """

    tff = FieldBased.from_param(tff).field

    ivtc = clip.std.SeparateFields(tff=tff).std.DoubleWeave()
    ivtc = IVTCycles.cycle_10.decimate(ivtc, pattern)

    return FieldBased.PROGRESSIVE.apply(ivtc)


def jivtc(
    src: vs.VideoNode, pattern: int, tff: bool = True, chroma_only: bool = True,
    postprocess: VSFunctionKwArgs = deblend, postdecimate: bool = True, **kwargs: Any  # type: ignore
) -> vs.VideoNode:
    """
    This function should only be used when a normal ivtc or ivtc + bobber leaves chroma blend to a every fourth frame.
    You can disable chroma_only to use in luma as well, but it is not recommended.

    :param src:             Source clip. Has to be 60i.
    :param pattern:         First frame of any clean-combed-combed-clean-clean sequence.
    :param tff:             Set top field first (True) or bottom field first (False).
    :param chroma_only:     Decide whether luma too will be processed.
    :param postprocess:     Function to run after second decimation. Should be either a bobber or a deblender.
    :param postdecimate:    If the postprocess function doesn't decimate itself, put True.

    :return:            Inverse Telecined clip.
    """

    InvalidFramerateError.check(jivtc, src, (30000, 1001))

    ivtced = core.std.SeparateFields(src, tff=tff).std.DoubleWeave()
    ivtced = IVTCycles.cycle_10.decimate(ivtced, pattern)

    pprocess = postprocess(src if postdecimate else ivtced, **kwargs)

    if postdecimate:
        pprocess = IVTCycles.cycle_05.decimate(pprocess, pattern)

    inter = core.std.Interleave([ivtced, pprocess])
    final = IVTCycles.cycle_08.decimate(inter, pattern)

    final = join(ivtced, final) if chroma_only else final

    return FieldBased.ensure_presence(final, FieldBased.PROGRESSIVE)


main_file = os.path.realpath(sys.argv[0]) if sys.argv[0] else None
main_file = os.path.splitext(os.path.basename(str(main_file)))[0]
main_file = "{yourScriptName}_" if main_file in ("__main___", "setup_") else main_file


def tivtc_2pass(
    clip: vs.VideoNode,
    decimate_mode: int | tuple[int, int], decimate: int | bool = True,
    tfm_in: Path | str = f".ivtc/{main_file}_matches.txt",
    tdec_in: Path | str = f".ivtc/{main_file}_metrics.txt",
    timecodes_out: Path | str = f".ivtc/{main_file}_timecodes.txt",
    tfm_args: dict[str, Any] | None = None, tdecimate_args: dict[str, Any] | None = None,
    tfm_pass_args: tuple[dict[str, Any] | None, dict[str, Any] | None] | None = None,
    tdecimate_pass_args: tuple[dict[str, Any] | None, dict[str, Any] | None] | None = None,
) -> vs.VideoNode:
    """
    Perform TFM and TDecimate on a clip.

    Includes automatic generation of a metrics/matches/timecodes txt file.

    Dependencies:

    * `TIVTC <https://github.com/dubhater/vapoursynth-tivtc>`_


    :param clip:                Clip to process.
    :param decimate_mode:       Decimate mode for both passes of TDecimate.
    :param decimate:            Whether to perform TDecimate on the clip or returns TFM'd clip only.
                                Set to -1 to use TDecimate without TFM.
    :param tfm_in:              Location for TFM's matches analysis.
    :param tdec_in:             Location for TDecimate's metrics analysis.
    :param timecodes_out:       Location for TDecimate's timecodes analysis.
    :param tfm_args:            Additional arguments to pass to TFM.
    :param tdecimate_args:      Additional arguments to pass to TDecimate.
    :param tfm_pass_args:       Arguments that will overwrite ``tfm_args`` in the first and second pass.
    :param tdecimate_pass_args: Arguments that will overwrite ``tdecimate_args`` in the first and second pass.

    :return:                    IVTC'd clip with external timecode/matches/metrics txt files.

    :raises TypeError:          Invalid ``decimate`` argument is passed.
    """

    decimate = int(decimate)

    if decimate not in {-1, 0, 1}:
        raise CustomTypeError(
            "Invalid 'decimate' argument. Must be True/False, their integer values, or -1!", tivtc_vfr
        )

    tfm_args, tdecimate_args = tfm_args or {}, tdecimate_args or {}

    if tfm_pass_args:
        tfm_pass1, tfm_pass2 = tuple((tfm_args | (x or {})) for x in tfm_pass_args)
    else:
        tfm_pass1, tfm_pass2 = tfm_args.copy(), tfm_args.copy()

    if tdecimate_pass_args:
        tdecimate_pass1, tdecimate_pass2 = tuple((tdecimate_args | (x or {})) for x in tdecimate_pass_args)
    else:
        tdecimate_pass1, tdecimate_pass2 = tdecimate_args.copy(), tdecimate_args.copy()

    if isinstance(decimate_mode, tuple):
        tdec_mode1, tdec_mode2 = decimate_mode
    else:
        tdec_mode1 = tdec_mode2 = decimate_mode

    tdecimate_pass1, tdecimate_pass2 = tdecimate_pass1 | dict(mode=tdec_mode1), tdecimate_pass2 | dict(mode=tdec_mode2)

    tfm_f = tdec_f = timecodes_f = Path()

    def _set_paths() -> None:
        nonlocal tfm_f, tdec_f, timecodes_f
        tfm_f = Path(tfm_in).resolve().absolute()
        tdec_f = Path(tdec_in).resolve().absolute()
        timecodes_f = Path(timecodes_out).resolve().absolute()

    _set_paths()

    for p in (tfm_f, tdec_f, timecodes_f):
        p.parent.mkdir(parents=True, exist_ok=True)
        if p.exists() and p.stat().st_size == 0:
            p.unlink(True)

    if not (tfm_f.exists() and tdec_f.exists()):
        ivtc_clip = clip.tivtc.TFM(output=str(tfm_f), **tfm_pass1)
        ivtc_clip = ivtc_clip.tivtc.TDecimate(**tdecimate_pass1, output=str(tdec_f))

        clip_async_render(ivtc_clip, None, "Calculating matches and metrics...")

        ivtc_clip = None  # type: ignore
        del ivtc_clip

        gc.collect(0)
        gc.collect(1)
        gc.collect(2)

        _set_paths()

    if decimate != -1:
        clip = clip.tivtc.TFM(**tfm_pass2, input=str(tfm_f))

    if decimate == 0:
        return clip

    return clip.tivtc.TDecimate(
        **tdecimate_pass2, input=str(tdec_f), tfmIn=str(tfm_f), mkvOut=str(timecodes_f)
    )


def tivtc_vfr(
    clip: vs.VideoNode,
    tfm_in: Path | str = f".ivtc/{main_file}_matches.txt",
    tdec_in: Path | str = f".ivtc/{main_file}_metrics.txt",
    timecodes_out: Path | str = f".ivtc/{main_file}_timecodes.txt",
    decimate: int | bool = True, tfm_args: dict[str, Any] | None = None,
    tdecimate_args: dict[str, Any] | None = None, **kwargs: Any
) -> vs.VideoNode:
    """
    Perform TFM and TDecimate on a clip that is supposed to be VFR.

    Includes automatic generation of a metrics/matches/timecodes txt file.

    Dependencies:

    * `TIVTC <https://github.com/dubhater/vapoursynth-tivtc>`_


    :param clip:                Clip to process.
    :param tfm_in:              Location for TFM's matches analysis.
    :param tdec_in:             Location for TDecimate's metrics analysis.
    :param timecodes_out:       Location for TDecimate's timecodes analysis.
    :param decimate:            Whether to perform TDecimate on the clip or returns TFM'd clip only.
                                Set to -1 to use TDecimate without TFM.
    :param tfm_args:            Additional arguments to pass to TFM.
    :param tdecimate_args:      Additional arguments to pass to TDecimate.
    :param kwargs:              Additional kwargs for ``tivtc_2pass``.

    :return:                    IVTC'd VFR clip with external timecode/matches/metrics txt files.

    :raises TypeError:          Invalid ``decimate`` argument is passed.
    """

    return tivtc_2pass(
        clip, (4, 5), decimate, tfm_in, tdec_in, timecodes_out, tfm_args,
        tdecimate_args, None, (None, dict(hybrid=2, vfrDec=1)), **kwargs
    )



from __future__ import annotations

from typing import cast

from vsexprtools import ExprVars, complexpr_available, norm_expr
from vsrgtools import sbr
from vstools import (
    ConvMode, CustomEnum, FieldBased, FieldBasedT, FuncExceptT, FunctionUtil, PlanesT, core, depth, expect_bits,
    get_neutral_values, scale_8bit, vs
)

__all__ = [
    'fix_telecined_fades',

    'vinverse'
]


def fix_telecined_fades(
    clip: vs.VideoNode, tff: bool | FieldBasedT | None = None, colors: float | list[float] = 0.0,
    planes: PlanesT = None, func: FuncExceptT | None = None
) -> vs.VideoNode:
    """
    Give a mathematically perfect solution to fades made *after* telecining (which made perfect IVTC impossible).

    This is an improved version of the Fix-Telecined-Fades plugin.

    Make sure to run this *after* IVTC/deinterlacing!

    :param clip:                            Clip to process.
    :param tff:                             Top-field-first. `False` sets it to Bottom-Field-First.
                                            If `None`, get the field order from the _FieldBased prop.
    :param colors:                          Color offset for the plane average.

    :return:                                Clip with fades (and only fades) accurately deinterlaced.

    :raises UndefinedFieldBasedError:       No automatic ``tff`` can be determined.
    """
    func = func or fix_telecined_fades

    if not complexpr_available:
        raise ExprVars._get_akarin_err()(func=func)

    clip = FieldBased.ensure_presence(clip, tff, func)

    f = FunctionUtil(clip, func, planes, (vs.GRAY, vs.YUV), 32)

    fields = f.work_clip.std.Limiter().std.SeparateFields()

    for i in f.norm_planes:
        fields = fields.std.PlaneStats(None, i, f'P{i}')

    props_clip = core.akarin.PropExpr(
        [f.work_clip, fields[::2], fields[1::2]], lambda: {  # type: ignore[misc]
            f'f{t}Avg{i}': f'{c}.P{i}Average {color} -'  # type: ignore[has-type]
            for t, c in ['ty', 'bz']
            for i, color in zip(f.norm_planes, f.norm_seq(colors))
        }
    )

    fix = norm_expr(
        props_clip, 'Y 2 % BF! BF@ x.fbAvg{i} x.ftAvg{i} ? AVG! '
        'AVG@ 0 = x x {color} - AVG@ BF@ x.ftAvg{i} x.fbAvg{i} ? + 2 / AVG@ / * ? {color} +',
        planes, i=f.norm_planes, color=colors, force_akarin=func
    )

    return f.return_clip(fix)


class Vinverse(CustomEnum):
    V1 = object()
    V2 = object()
    MASKED = object()
    MASKEDV1 = object()
    MASKEDV2 = object()

    def __call__(
        self, clip: vs.VideoNode, sstr: float = 2.7, amount: int = 255, scale: float = 0.25,
        mode: ConvMode = ConvMode.VERTICAL, planes: PlanesT = None
    ) -> vs.VideoNode:
        if amount <= 0:
            return clip

        clip, bits = expect_bits(clip, 32)

        neutral = get_neutral_values(clip)

        expr = f'y z - {sstr} * D1! x y - D2! D1@ abs D1A! D2@ abs D2A! '
        expr += f'D1@ D2@ xor D1A@ D2A@ < D1@ D2@ ? {scale} * D1A@ D2A@ < D1@ D2@ ? ? y + '

        if self in {Vinverse.V1, Vinverse.MASKEDV1}:
            blur = clip.std.Convolution([50, 99, 50], mode=mode, planes=planes)
            blur2 = blur.std.Convolution([1, 4, 6, 4, 1], mode=mode, planes=planes)
        elif self in {Vinverse.V2, Vinverse.MASKEDV2}:
            blur = sbr(clip, mode=mode, planes=planes)
            blur2 = blur.std.Convolution([1, 2, 1], mode=mode, planes=planes)

        if self in {Vinverse.MASKED, Vinverse.MASKEDV1, Vinverse.MASKEDV2}:
            search_str = 'x[-1,0] x[1,0]' if mode == ConvMode.HORIZONTAL else 'x[0,-1] x[0,1]'
            mask_search_str = search_str.replace('x', 'y')

            if self is Vinverse.MASKED:
                find_combs = norm_expr(clip, f'x x 2 * {search_str} + + 4 / - {{n}} +', planes, n=neutral)
                decomb = norm_expr(
                    [find_combs, clip],
                    'x x 2 * {search_str} + + 4 / - B! y B@ x {n} - * 0 '
                    '< {n} B@ abs x {n} - abs < B@ {n} + x ? ? - {n} +', n=neutral, search_str=search_str
                )
            else:
                decomb = norm_expr(
                    [clip, blur, blur2], 'x x 2 * y + 4 / - {n} + FC! FC@ FC@ 2 * y z - {n} + + 4 / - B! '
                    'x B@ FC@ {n} - * 0 < {n} B@ abs FC@ {n} - abs < B@ {n} + FC@ ? ? - {n} +', n=neutral
                )

            return norm_expr(
                [clip, decomb], f'{scale_8bit(clip, amount)} a! y y y y 2 * {mask_search_str} + + 4 / - {sstr} '
                '* + y - {n} + D1! x y - {n} + D2! D1@ {n} - D2@ {n} - * 0 < D1@ {n} - abs D2@ {n} - abs < D1@ '
                'D2@ ? {n} - {scale} * {n} + D1@ {n} - abs D2@ {n} - abs < D1@ D2@ ? ? {n} - + merge! '
                'x a@ + merge@ < x a@ + x a@ - merge@ > x a@ - merge@ ? ?', n=neutral, scale=scale
            )

        if amount < 255:
            amn = scale_8bit(clip, amount)
            expr += f'LIM! x {amn} + LIM@ < x {amn} + x {amn} - LIM@ > x {amn} - LIM@ ? ?'

        return depth(norm_expr([clip, blur, blur2], expr, planes), bits)


vinverse = cast(Vinverse, Vinverse.V1)



from __future__ import annotations

from typing import Any, cast

from vsexprtools import complexpr_available, expr_func, norm_expr
from vstools import VSFunction, join, shift_clip, shift_clip_multi, vs

from .combing import vinverse
from .utils import telecine_patterns

__all__ = [
    'deblending_helper',

    'deblend', 'deblend_bob',

    'deblend_fix_kf'
]


def deblending_helper(deblended: vs.VideoNode, fieldmatched: vs.VideoNode, length: int = 5) -> vs.VideoNode:
    """
    Helper function to select a deblended clip pattern from a fieldmatched clip.

    :param deblended:       Deblended clip.
    :param fieldmatched:    Source after field matching, must have field=3 and possibly low cthresh.
    :param length:          Length of the pattern.

    :return: Deblended clip.
    """
    inters = telecine_patterns(fieldmatched, deblended, length)
    inters += [shift_clip(inter, 1) for inter in inters]

    inters.insert(0, fieldmatched)

    prop_srcs = shift_clip_multi(fieldmatched, (0, 1))

    if complexpr_available:
        index_src = expr_func(
            prop_srcs, f'x._Combed N {length} % 1 + y._Combed {length} 0 ? + 0 ?', vs.GRAY8
        )

        return fieldmatched.std.FrameEval(lambda n, f: inters[f[0][0, 0]], index_src)  # type: ignore

    def _deblend_eval(n: int, f: list[vs.VideoFrame]) -> vs.VideoNode:
        idx = 0

        if f[0].props._Combed == 1:
            idx += (n % length) + 1

            if f[1].props._Combed == 1:
                idx += length

        return inters[idx]

    return fieldmatched.std.FrameEval(_deblend_eval, prop_srcs)


def deblend(
    src: vs.VideoNode, fieldmatched: vs.VideoNode | None = None, decomber: VSFunction | None = vinverse, **kwargs: Any
) -> vs.VideoNode:
    """
    Automatically deblends if normal field matching leaves 2 blends every 5 frames. Adopted from jvsfunc.

    :param src:             Input source to fieldmatching.
    :param fieldmatched:    Source after field matching, must have field=3 and possibly low cthresh.
    :param decomber:        Optional post processing decomber after deblending and before pattern matching.

    :return: Deblended clip.
    """

    deblended = norm_expr(shift_clip_multi(src, (-1, 2)), 'z a 2 / - y x 2 / - +')

    if decomber:
        deblended = decomber(deblended, **kwargs)

    if fieldmatched:
        deblended = deblending_helper(fieldmatched, deblended)

    return join(fieldmatched or src, deblended)


def deblend_bob(
    bobbed: vs.VideoNode | tuple[vs.VideoNode, vs.VideoNode],
    fieldmatched: vs.VideoNode | None = None, blend_out: bool = False
) -> vs.VideoNode:
    """
    Stronger version of `deblend` that uses a bobbed clip to deblend. Adopted from jvsfunc.

    :param bobbed:          Bobbed source or a tuple of even/odd fields.
    :param fieldmatched:    Source after field matching, must have field=3 and possibly low cthresh.

    :return: Deblended clip.
    """

    if isinstance(bobbed, tuple):
        bob0, bob1 = bobbed
    else:
        bob0, bob1 = bobbed.std.SelectEvery(2, 0), bobbed.std.SelectEvery(2, 1)

    ab0, bc0, c0 = shift_clip_multi(bob0, (0, 1, 2))
    bc1, ab1, a1 = shift_clip_multi(bob1)

    deblended = norm_expr([a1, ab1, ab0, bc1, bc0, c0], ('b', 'y x - z + b c - a + + 2 /'))

    if fieldmatched:
        return deblending_helper(fieldmatched, deblended)

    return deblended


def deblend_fix_kf(deblended: vs.VideoNode, fieldmatched: vs.VideoNode) -> vs.VideoNode:
    """
    Should be used after deblend/_bob to fix scene changes. Adopted from jvsfunc.

    :param deblended:       Deblended clip.
    :param fieldmatched:    Fieldmatched clip used to debled, must have field=3 and possibly low cthresh.

    :return: Deblended clip with fixed blended keyframes.
    """

    shifted_clips = shift_clip_multi(deblended)
    prop_srcs = shift_clip_multi(fieldmatched, (0, 1))

    if complexpr_available:
        index_src = expr_func(
            prop_srcs, 'x._Combed x.VFMSceneChange and y.VFMSceneChange 2 0 ? 1 ?', vs.GRAY8
        )

        return deblended.std.FrameEval(lambda n, f: shifted_clips[f[0][0, 0]], index_src)  # type: ignore

    def _keyframe_fix(n: int, f: list[vs.VideoFrame]) -> vs.VideoNode:
        keyfm = cast(tuple[int, int], (f[0].props.VFMSceneChange, f[1].props.VFMSceneChange))

        idx = 1
        if f[0].props._Combed == 1:
            if keyfm == (1, 0):
                idx = 0
            elif keyfm == (1, 1):
                idx = 2

        return shifted_clips[idx]

    return deblended.std.FrameEval(_keyframe_fix, prop_srcs)
